(function() {
    "use strict";
    function QueryComparisons() {}
    QueryComparisons.prototype = {
        $default: function(modelValue, testValue) {
            if (testValue instanceof RegExp) return this.compareMethods.$regexp(modelValue, testValue); else return this.compareMethods.$equals(modelValue, testValue);
        },
        $equals: function(modelValue, testValue) {
            return modelValue === testValue;
        },
        $length: function(modelValue, testValue) {
            return modelValue.length == testValue;
        },
        $regexp: function(modelValue, testValue) {
            return testValue.test(modelValue);
        },
        $left: function(modelValue, testValue) {
            return modelValue.substring(0, testValue.length) === testValue;
        },
        $right: function(modelValue, testValue) {
            var offset = modelValue.length - testValue.length;
            return modelValue.substring(offset) === testValue;
        },
        $typeof: function(modelValue, testValue) {
            return typeof modelValue === testValue;
        },
        $instanceof: function(modelValue, testValue) {
            return modelValue instanceof testValue;
        },
        $toLowerCase: function(modelValue, testValue) {
            return modelValue.toLowerCase() === testValue;
        },
        $toUpperCase: function(modelValue, testValue) {
            return modelValue.toUpperCase() === testValue;
        },
        $inArray: function(modelValue, testArray) {
            return testArray.indexOf(modelValue) !== -1;
        },
        $has: function(modelValue, testValue) {
            if (testValue instanceof Array) {
                return testValue.every(function(testItemValue) {
                    return testItemValue in modelValue;
                });
            } else {
                return testValue in modelValue;
            }
        },
        $g: function(modelValue, testValue) {
            return modelValue > testValue;
        },
        $gte: function(modelValue, testValue) {
            return modelValue >= testValue;
        },
        $lte: function(modelValue, testValue) {
            return modelValue <= testValue;
        },
        $l: function(modelValue, testValue) {
            return modelValue < testValue;
        },
        $e: function(modelValue, testValue) {
            return modelValue == testValue;
        },
        $ne: function(modelValue, testValue) {
            return modelValue != testValue;
        },
        $queryOne: function(modelValue, criteria) {
            return this.opMethods.executeQuery.call(this, modelValue, criteria, false);
        },
        $queryEvery: function(modelArray, criteria) {
            var scope = this;
            return modelArray.every(function(model) {
                return scope.opMethods.executeQuery.call(scope, model, criteria, false);
            });
        },
        $querySome: function(modelArray, criteria) {
            var scope = this;
            return modelArray.some(function(model) {
                return scope.opMethods.executeQuery.call(scope, model, criteria, false);
            });
        }
    };
    function QueryOperations() {}
    function throwError(error) {
        throw error;
    }
    QueryOperations.prototype = {
        $and: function(model, operations, negateCompare, modelValue) {
            if (this.trace.on === true) {
                this.trace.write("$and");
            }
            return this.opMethods.executeQuery.call(this, model, operations, negateCompare, modelValue);
        },
        $or: function(model, operations, negateCompare, modelValue) {
            if (this.trace.on === true) {
                this.trace.write("$or");
            }
            return this.opMethods.executeQuery.call(this, model, operations, negateCompare, modelValue);
        },
        $not: function(model, operations, modelValue) {
            var stackIndex;
            if (this.trace.on === true) {
                stackIndex = this.trace.stackLevel;
                this.trace.write("$not");
            }
            var result = this.opMethods.executeQuery.call(this, model, operations, true, modelValue);
            if (this.trace.on === true) {
                this.trace.stack[stackIndex - 1].result = result;
            }
            return result;
        },
        executeCompare: function(operationkey, operations, negateCompare, modelValue, reject) {
            if (this.trace.on === true) {
                this.trace.write(operationkey);
            }
            var compareMethod = this.compareMethods[operationkey];
            if (compareMethod === undefined) {
                reject(new ReferenceError("No compare method was found for: " + operationkey));
                return;
            }
            var compareValue = operations[operationkey];
            var result = compareMethod.call(this, modelValue, compareValue);
            if (negateCompare) {
                result = !result;
            }
            if (this.trace.on === true) {
                var traceStack = this.trace.stack;
                traceStack[traceStack.length - 1].result = result;
            }
            return result;
        },
        executeValueOperations: function(model, operationkey, operations, negateCompare, reject) {
            if (this.trace.on === true) {
                this.trace.write(operationkey);
            }
            var descendantOps = operations[operationkey];
            if (descendantOps === undefined) {
                reject(new ReferenceError("No operation property was found for: " + operationkey));
                return;
            }
            var modelValue = model[operationkey];
            return this.opMethods.executeQuery.call(this, model, descendantOps, negateCompare, modelValue);
        },
        executeQuery: function(model, operations, negateCompare, modelValue) {
            var context = this;
            if (context.trace.on === true) {
                context.trace.stackLevel++;
            }
            if (operations instanceof RegExp || typeof operations !== "object") {
                operations = {
                    $default: operations
                };
            }
            var result = false;
            for (var operationKey in operations) {
                if (operations.hasOwnProperty(operationKey)) {
                    if (operationKey[0] === "$") {
                        if (modelValue === undefined) {
                            modelValue = model;
                        }
                        if (operationKey === "$not") {
                            result = context.opMethods.$not.call(context, model, operations.$not, modelValue);
                        } else if (operationKey === "$and") {
                            result = context.opMethods.$and.call(context, model, operations.$and, negateCompare, modelValue);
                        } else if (operationKey !== "$or") {
                            result = context.opMethods.executeCompare.call(context, operationKey, operations, negateCompare, modelValue, throwError);
                        }
                    } else {
                        result = context.opMethods.executeValueOperations.call(context, model, operationKey, operations, negateCompare, throwError);
                    }
                    if (result === false) break;
                }
            }
            if (operations.$or && result === false) {
                result = context.opMethods.$or.call(context, model, operations.$or, negateCompare, modelValue);
            }
            if (context.trace.on === true) {
                context.trace.stackLevel--;
            }
            return result;
        }
    };
    function QueryPromise(queryOperations, queryComparisons, queryStackTrace, sharedContext) {
        this.opMethods = queryOperations;
        this.compareMethods = queryComparisons;
        this.trace = queryStackTrace;
        this.sharedContext = sharedContext || {};
    }
    QueryPromise.prototype = {
        extract: function(modelDictionary, criteria) {
            var context = this, trace = context.trace.on;
            if (trace === true) {
                context.trace.reset();
            }
            if (modelDictionary === undefined) {
                return Promise.reject(new ReferenceError("Unspecified model dictionary."));
            }
            if (criteria === undefined) {
                return Promise.reject(new ReferenceError("Unspecified criteria."));
            }
            var chainedThen = Promise.resolve({});
            var keys = Object.keys(modelDictionary);
            keys.forEach(function(modelKey, modelIndex) {
                chainedThen = chainedThen.then(function(filteredDictionary) {
                    if (trace === true) {
                        context.trace.stackLevel = 0;
                        context.trace.write([ "$$extract", modelIndex ]);
                    }
                    var model = modelDictionary[modelKey];
                    return context.opMethods.executeQuery.call(context, model, criteria, false).then(function(result) {
                        if (result === true) {
                            filteredDictionary[modelKey] = model;
                        }
                        return filteredDictionary;
                    });
                });
            });
            return chainedThen;
        },
        filter: function(modelArray, criteria) {
            var context = this, trace = context.trace.on;
            if (trace === true) {
                context.trace.reset();
            }
            if (modelArray === undefined) {
                return Promise.reject(new ReferenceError("Unspecified model array."));
            }
            if (criteria === undefined) {
                return Promise.reject(new ReferenceError("Unspecified criteria."));
            }
            var chainedThen = Promise.resolve([]);
            modelArray.forEach(function(model, modelIndex) {
                chainedThen = chainedThen.then(function(filteredList) {
                    if (trace === true) {
                        context.trace.stackLevel = 0;
                        context.trace.write([ "$$filter", modelIndex ]);
                    }
                    return context.opMethods.executeQuery.call(context, model, criteria, false).then(function(result) {
                        if (result === true) {
                            filteredList.push(model);
                        }
                        return filteredList;
                    });
                });
            });
            return chainedThen;
        },
        every: function(modelArray, criteria) {
            var context = this, trace = context.trace.on;
            if (trace === true) {
                context.trace.reset();
            }
            if (modelArray === undefined) {
                return Promise.reject(new ReferenceError("Unspecified model array."));
            }
            if (criteria === undefined) {
                return Promise.reject(new ReferenceError("Unspecified criteria."));
            }
            return promiseExtras.every(modelArray, function(model, modelIndex, fulfil) {
                if (trace === true) {
                    context.trace.stackLevel = 0;
                    context.trace.write([ "$$every", modelIndex ]);
                }
                fulfil(context.opMethods.executeQuery.call(context, model, criteria, false));
            });
        },
        one: function(model, criteria) {
            if (this.trace.on === true) {
                this.trace.reset();
            }
            if (model === undefined) {
                return Promise.reject(new ReferenceError("Unspecified model."));
            }
            if (criteria === undefined) {
                return Promise.reject(new ReferenceError("Unspecified criteria."));
            }
            return this.opMethods.executeQuery.call(this, model, criteria, false);
        },
        some: function(modelArray, criteria) {
            var context = this, trace = context.trace.on;
            if (trace === true) {
                context.trace.reset();
            }
            if (modelArray === undefined) {
                return Promise.reject(new ReferenceError("Unspecified model array."));
            }
            if (criteria === undefined) {
                return Promise.reject(new ReferenceError("Unspecified criteria."));
            }
            return promiseExtras.some(modelArray, function(model, modelIndex, fulfil) {
                if (trace === true) {
                    context.trace.stackLevel = 0;
                    context.trace.write([ "$$some", modelIndex ]);
                }
                fulfil(context.opMethods.executeQuery.call(context, model, criteria, false));
            });
        }
    };
    function QueryPromiseComparisons() {}
    var queryComparisonsProto = Object.create(QueryComparisons.prototype);
    queryComparisonsProto.$queryEvery = function(modelArray, criteria) {
        var scope = this;
        return promiseExtras.every(modelArray, function(model, modelIndex, fulfil) {
            fulfil(scope.opMethods.executeQuery.call(scope, model, criteria, false));
        });
    };
    queryComparisonsProto.$querySome = function(modelArray, criteria) {
        var scope = this;
        return promiseExtras.some(modelArray, function(model, modelIndex, fulfil) {
            fulfil(scope.opMethods.executeQuery.call(scope, model, criteria, false));
        });
    };
    QueryPromiseComparisons.prototype = queryComparisonsProto;
    function QueryPromiseOperations() {}
    var queryOperationsProto = Object.create(QueryOperations.prototype);
    queryOperationsProto.executeQuery = function(model, operations, negateCompare, modelValue) {
        var context = this, trace = context.trace.on, operationKeys, hasOrOperation;
        if (trace === true) {
            context.trace.stackLevel++;
        }
        if (operations instanceof RegExp || typeof operations !== "object") {
            operations = {
                $default: operations
            };
        }
        operationKeys = Object.keys(operations);
        hasOrOperation = "$or" in operations;
        if (hasOrOperation === true) {
            operationKeys = operationKeys.filter(function(element) {
                return element !== "$or";
            });
        }
        return promiseExtras.every(operationKeys, function(operationkey, operationIndex, fulfil, reject) {
            var result = false;
            if (operationkey[0] === "$") {
                if (modelValue === undefined) {
                    modelValue = model;
                }
                if (operationkey === "$not") {
                    result = context.opMethods.$not.call(context, model, operations.$not, modelValue);
                } else if (operationkey === "$and") {
                    result = context.opMethods.$and.call(context, model, operations.$and, negateCompare, modelValue);
                } else {
                    result = context.opMethods.executeCompare.call(context, operationkey, operations, negateCompare, modelValue, reject);
                }
            } else {
                result = context.opMethods.executeValueOperations.call(context, model, operationkey, operations, negateCompare, reject);
            }
            fulfil(result);
        }).then(function(result) {
            if (operationKeys.length === 0) {
                result = false;
            }
            if (hasOrOperation && result === false) {
                result = context.opMethods.$or.call(context, model, operations.$or, negateCompare, modelValue);
            }
            return result;
        }).then(function(result) {
            if (trace === true) {
                context.trace.stackLevel--;
            }
            return result;
        });
    };
    QueryPromiseOperations.prototype = queryOperationsProto;
    function QueryStackTrace() {
        this.on = false;
        this.reset();
    }
    QueryStackTrace.prototype = {
        reset: function() {
            this.stack = [];
            this.stackLevel = 0;
        },
        write: function(operation) {
            var newStackEntry = {
                level: this.stackLevel,
                operation: operation
            };
            this.stack.push(newStackEntry);
        }
    };
    function Query(queryOperations, queryComparisons, queryStackTrace, sharedContext) {
        this.opMethods = queryOperations;
        this.compareMethods = queryComparisons;
        this.trace = queryStackTrace;
        this.sharedContext = sharedContext || {};
    }
    Query.prototype = {
        extract: function(modelDictionary, criteria) {
            var trace = this.trace.on, filteredDictionary = {}, keyIndex, modelKey, model;
            if (trace === true) {
                this.trace.reset();
            }
            if (modelDictionary === undefined) {
                throw new ReferenceError("Unspecified model dictionary.");
            }
            if (criteria === undefined) {
                throw new ReferenceError("Unspecified criteria.");
            }
            var modelDictKeys = Object.keys(modelDictionary);
            for (keyIndex = 0; keyIndex < modelDictKeys.length; keyIndex++) {
                if (trace === true) {
                    this.trace.stackLevel = 0;
                    this.trace.write([ "$$extract", modelKey ]);
                }
                modelKey = modelDictKeys[keyIndex];
                model = modelDictionary[modelKey];
                var result = this.opMethods.executeQuery.call(this, model, criteria, false);
                if (result === true) {
                    filteredDictionary[modelKey] = model;
                }
            }
            return filteredDictionary;
        },
        filter: function(modelArray, criteria) {
            var trace = this.trace.on;
            if (trace === true) {
                this.trace.reset();
            }
            if (modelArray === undefined) {
                throw new ReferenceError("Unspecified model array.");
            }
            if (criteria === undefined) {
                throw new ReferenceError("Unspecified criteria.");
            }
            var result, model, traceIndex = 0, nextIter = modelArray[Symbol.iterator](), iterResult, filteredList = [];
            for (;!(iterResult = nextIter.next()).done; ) {
                if (trace === true) {
                    this.trace.stackLevel = 0;
                    this.trace.write([ "$$filter", traceIndex++ ]);
                }
                model = iterResult.value;
                result = this.opMethods.executeQuery.call(this, model, criteria, false);
                if (result === true) {
                    filteredList.push(model);
                }
            }
            return filteredList;
        },
        every: function(modelArray, criteria) {
            var trace = this.trace.on;
            if (trace === true) {
                this.trace.reset();
            }
            if (modelArray === undefined) {
                throw new ReferenceError("Unspecified model array.");
            }
            if (criteria === undefined) {
                throw new ReferenceError("Unspecified criteria.");
            }
            var result, traceIndex = 0, nextIter = modelArray[Symbol.iterator](), iterResult;
            for (;!(iterResult = nextIter.next()).done; ) {
                if (trace === true) {
                    this.trace.stackLevel = 0;
                    this.trace.write([ "$$every", traceIndex++ ]);
                }
                result = this.opMethods.executeQuery.call(this, iterResult.value, criteria, false);
                if (result === false) break;
            }
            return result;
        },
        one: function(model, criteria) {
            if (this.trace.one === true) {
                this.trace.reset();
            }
            if (model === undefined) {
                throw new ReferenceError("Unspecified model.");
            }
            if (criteria === undefined) {
                throw new ReferenceError("Unspecified criteria.");
            }
            return this.opMethods.executeQuery.call(this, model, criteria, false);
        },
        some: function(modelArray, criteria) {
            var context = this, trace = context.trace.on;
            if (trace === true) {
                this.trace.reset();
            }
            if (modelArray === undefined) {
                throw new ReferenceError("Unspecified model array.");
            }
            if (criteria === undefined) {
                throw new ReferenceError("Unspecified criteria.");
            }
            var result, traceIndex = 0, nextIter = modelArray[Symbol.iterator](), iterResult;
            for (;!(iterResult = nextIter.next()).done; ) {
                if (trace === true) {
                    this.trace.stackLevel = 0;
                    this.trace.write([ "$$some", traceIndex++ ]);
                }
                result = this.opMethods.executeQuery.call(this, iterResult.value, criteria, false);
                if (result === true) break;
            }
            return result;
        }
    };
    var counter = 0;
    function newUniqueString() {
        return "__$" + Math.floor(Math.random() * 1e9) + "$" + counter++ + "$__";
    }
    var globalObj = (0, eval)("this");
    if (typeof Symbol === "undefined") {
        globalObj.Symbol = Object.create(null);
        if (Symbol.iterator === undefined) Symbol.iterator = newUniqueString();
    }
    function ArrayValuesIterator(arrayItems) {
        this.iteratedNextIndex = 0;
        this.iteratedObject = arrayItems;
    }
    ArrayValuesIterator.prototype = {
        next: function() {
            var done, value;
            done = this.iteratedNextIndex >= this.iteratedObject.length;
            if (done === false) value = this.iteratedObject[this.iteratedNextIndex++];
            return {
                value: value,
                done: done
            };
        }
    };
    ArrayValuesIterator.prototype[Symbol.iterator] = function() {
        return this;
    };
    function ArrayEntriesIterator(arrayItems) {
        this.iteratedNextIndex = 0;
        this.iteratedObject = arrayItems;
    }
    ArrayEntriesIterator.prototype = {
        superNext_: ArrayValuesIterator.prototype.next,
        next: function() {
            var value, index = this.iteratedNextIndex, nextResult = this.superNext_(), done = nextResult.done;
            if (done === false) value = [ index, nextResult.value ];
            return {
                value: value,
                done: done
            };
        }
    };
    ArrayEntriesIterator.prototype[Symbol.iterator] = function() {
        return this;
    };
    var arrayProto = Array.prototype;
    if (arrayProto.values === undefined) {
        arrayProto.values = function() {
            return new ArrayValuesIterator(this);
        };
    }
    if (arrayProto.entries === undefined) {
        arrayProto.entries = function() {
            return new ArrayEntriesIterator(this);
        };
    }
    if (arrayProto[Symbol.iterator] === undefined) {
        Object.defineProperty(arrayProto, Symbol.iterator, {
            value: arrayProto.values,
            configurable: true,
            enumerable: false,
            writable: true
        });
    }
    var isBrowser = typeof window !== typeof undefined, $Promise, $promiseExtras;
    function resolveBrowserPromiseLib() {
        if (window.Promise) {
            return window.Promise;
        } else {
            throw new ReferenceError("Promise library could not be detected for this browser.");
        }
    }
    function resolveBrowserPromiseExtrasLib() {
        if (window.promiseExtras) {
            return window.promiseExtras;
        } else {
            throw new ReferenceError("Promise-extras library could not be detected for this browser.");
        }
    }
    function lazyLoadPromiseDeps() {
        if (isBrowser === true) {
            $Promise = resolveBrowserPromiseLib();
        } else {
            $Promise = require("promise");
        }
        if (isBrowser === true) {
            $promiseExtras = resolveBrowserPromiseExtrasLib();
        } else {
            $promiseExtras = require("promise-extras");
        }
    }
    var defaultPromiseInstance;
    var queryifyLibrary = {
        get promise() {
            if (defaultPromiseInstance === undefined) {
                lazyLoadPromiseDeps();
                defaultPromiseInstance = new QueryPromise(new QueryPromiseOperations(), new QueryPromiseComparisons(), new QueryStackTrace());
            }
            return defaultPromiseInstance;
        },
        sync: new Query(new QueryOperations(), new QueryComparisons(), new QueryStackTrace()),
        Query: Query,
        QueryOperations: QueryOperations,
        QueryComparisons: QueryComparisons,
        QueryStackTrace: QueryStackTrace,
        get QueryPromise() {
            if (defaultPromiseInstance === undefined) lazyLoadPromiseDeps();
            return QueryPromise;
        },
        get QueryPromiseOperations() {
            if (defaultPromiseInstance === undefined) lazyLoadPromiseDeps();
            return QueryPromiseOperations;
        },
        get QueryPromiseComparisons() {
            if (defaultPromiseInstance === undefined) lazyLoadPromiseDeps();
            return QueryPromiseComparisons;
        },
        get polyfills() {
            return {
                ArrayValuesIterator: ArrayValuesIterator,
                ArrayEntriesIterator: ArrayEntriesIterator
            };
        }
    };
    if (typeof module !== typeof undefined && typeof module.exports !== typeof undefined) {
        module.exports = queryifyLibrary;
    } else {
        var indirectEval = (0, eval), globalThis = indirectEval("this");
        globalThis.queryify = queryifyLibrary;
    }
})();