/*
 QueryifyJs is a No-SQL object query library made for javascript
 Copyright (C) 2014 Peter Flannery

 This program is free software: you can redistribute it and/or modify
 it under the terms of the GNU Affero General Public License as
 published by the Free Software Foundation, either version 3 of the
 License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU Affero General Public License for more details.

 You should have received a copy of the GNU Affero General Public License
 along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/**
 * @module QueryifyJs
 */
/**
 * Promise based compare methods.
 *
 * @class QueryPromiseOperations
 * @extends QueryOperations
 */
function QueryPromiseOperations() {
}

var queryOperationsProto = Object.create(QueryOperations.prototype);

queryOperationsProto.executeQuery = function(model, operations, negateCompare, modelValue) {
  var context = this,
    trace = context.trace.on,
    operationKeys,
    hasOrOperation;

  if (trace === true) {
    // increment the query level
    context.trace.stackLevel++;
  }

  // use default compare if operations is not an object
  if (operations instanceof RegExp || typeof operations !== 'object') {
    operations = {
      '$default': operations
    };
  }

  // get the operation keys
  operationKeys = Object.keys(operations);

  // pop the $or if exists
  hasOrOperation = '$or' in operations;
  if (hasOrOperation === true) {
    operationKeys = operationKeys.filter(function(element) {
      return element !== '$or';
    })
  }

  // loop the operation keys
  return promiseExtras.every(operationKeys, function(operationkey, operationIndex, fulfil, reject) {
    var result = false;

    if (operationkey[0] === '$') {

      if (modelValue === undefined) {
        modelValue = model;
      }

      if (operationkey === '$not') {
        result = context.opMethods.$not.call(context, model, operations.$not, modelValue);
      } else if (operationkey === '$and') {
        result = context.opMethods.$and.call(context, model, operations.$and, negateCompare, modelValue);
      } else {
        result = context.opMethods.executeCompare.call(context,
          operationkey,
          operations,
          negateCompare,
          modelValue,
          reject
        );
      }

    } else {
      result = context.opMethods.executeValueOperations.call(context,
        model,
        operationkey,
        operations,
        negateCompare,
        reject
      );
    }

    fulfil(result);

  }).then(function(result) {
    // the condition where promiseExtras.every passes over an empty array and returns true
    if (operationKeys.length === 0) {
      result = false;
    }

    // run the $or if present and when the result is false
    if (hasOrOperation && result === false) {
      result = context.opMethods.$or.call(context, model, operations.$or, negateCompare, modelValue);
    }

    return result;

  }).then(function(result) {

    if (trace === true) {
      // decrement the query level
      context.trace.stackLevel--;
    }

    return result;

  });
};

QueryPromiseOperations.prototype = queryOperationsProto;
