/*
 QueryifyJs is a No-SQL object query library made for javascript
 Copyright (C) 2014 Peter Flannery

 This program is free software: you can redistribute it and/or modify
 it under the terms of the GNU Affero General Public License as
 published by the Free Software Foundation, either version 3 of the
 License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU Affero General Public License for more details.

 You should have received a copy of the GNU Affero General Public License
 along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/**
 * @module QueryifyJs
 */
/**
 * Useful for query diagnostics
 *
 * @class QueryStackTrace
 */
function QueryStackTrace() {
  /**
   * On\off switch for tracing operations and results made when using Query or QueryPromise methods
   *
   * @property on
   * @type Boolean
   *
   */
  this.on = false;

  this.reset();
}

QueryStackTrace.prototype = {
  /**
   * Resets the trace stack and level
   *
   * @method reset
   *
   */
  reset: function() {

    /**
     * Stores the trace stack of operations and results made by Query or QueryPromise
     *
     * @property stack
     * @type Array
     *
     */
    this.stack = [];
    this.stackLevel = 0;
  },
  /**
   * Writes to  the trace stack which stores the current level and operation parameter
   *
   * @method write
   *
   */
  write: function(operation) {
    var newStackEntry = {
      level: this.stackLevel,
      operation: operation
    };

    // store the new stack entry
    this.stack.push(newStackEntry);
  }
};
